@@ -6,27 +6,9 @@
  * @ref License
  */
 
-/* LICENSE
- *
- * Copyright (c) 2008-2019 OpenShot Studios, LLC
- * <http://www.openshotstudios.com/>. This file is part of
- * OpenShot Library (libopenshot), an open-source project dedicated to
- * delivering high quality video editing and animation solutions to the
- * world. For more information visit <http://www.openshot.org/>.
- *
- * OpenShot Library (libopenshot) is free software: you can redistribute it
- * and/or modify it under the terms of the GNU Lesser General Public License
- * as published by the Free Software Foundation, either version 3 of the
- * License, or (at your option) any later version.
- *
- * OpenShot Library (libopenshot) is distributed in the hope that it will be
- * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
- * GNU Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public License
- * along with OpenShot Library. If not, see <http://www.gnu.org/licenses/>.
- */
+// Copyright (c) 2008-2019 OpenShot Studios, LLC
+//
+// SPDX-License-Identifier: LGPL-3.0-or-later
 
 #include "Fraction.h"
 #include <cmath>
@@ -34,21 +16,20 @@
 using namespace openshot;
 
 // Delegating constructors
-Fraction::Fraction() : Fraction::Fraction(1, 1) {};
+Fraction::Fraction() : Fraction::Fraction(1, 1) {}
 
 Fraction::Fraction(std::pair<int, int> pair)
-	: Fraction::Fraction(pair.first, pair.second) {};
+	: Fraction::Fraction(pair.first, pair.second) {}
 
 Fraction::Fraction(std::map<std::string, int> mapping)
-	: Fraction::Fraction(mapping["num"], mapping["den"]) {};
+	: Fraction::Fraction(mapping["num"], mapping["den"]) {}
 
 Fraction::Fraction(std::vector<int> vector)
-	: Fraction::Fraction(vector[0], vector[1]) {};
+	: Fraction::Fraction(vector[0], vector[1]) {}
 
 // Full constructor
 Fraction::Fraction(int num, int den) :
-	num(num), den(den) {
-}
+	num(num), den(den) {}
 
 // Return this fraction as a float (i.e. 1/2 = 0.5)
 float Fraction::ToFloat() {
@@ -84,6 +65,9 @@ int Fraction::GreatestCommonDenominator() {
 void Fraction::Reduce() {
 	// Get the greatest common denominator
 	int GCD = GreatestCommonDenominator();
+	if (GCD == 0) {
+	    return;
+	}
 
 	// Reduce this fraction to the smallest possible whole numbers
 	num = num / GCD;
